#ifndef P2C_H
#define P2C_H

#include <stdio.h>

# include <stddef.h>
# include <stdlib.h>

#include <string.h>

#ifdef __linux__
#include <unistd.h>
#endif


#include <ctype.h>
#include <math.h>
#include <setjmp.h>
#include <assert.h>
#include <stdbool.h>

/* Function for compiler warnings*/
#ifdef __GNUC__
#define DEPRECATED __attribute__((deprecated))
#else
/**
 * Deprecated functions
 *
 * Use this macro in front of a function declaration to say the function is deprecated.
 * The compiler may emit a warning every time the function is used
 */
#define DEPRECATED
#endif


#define log1p(X) log(1+(X))

typedef struct __p2c_jmp_buf {
    struct __p2c_jmp_buf *next;
    jmp_buf jbuf;
} __p2c_jmp_buf;


/* Warning: The following will not work if setjmp is used simultaneously.
   This also violates the ANSI restriction about using vars after longjmp,
   but a typical implementation of longjmp will get it right anyway. */
#ifndef FAKE_TRY
# define TRY(x)         do { __p2c_jmp_buf __try_jb;  \
			     __try_jb.next = __top_jb;  \
			     if (!setjmp((__top_jb = &__try_jb)->jbuf)) {
# define RECOVER(x)	__top_jb = __try_jb.next; } else {
# define RECOVER2(x,L)  __top_jb = __try_jb.next; } else {  \
			     if (0) { L: __top_jb = __try_jb.next; }
# define ENDTRY(x)      } } while (0)
#else
# define TRY(x)         if (1) {
# define RECOVER(x)     } else do {
# define RECOVER2(x,L)  } else do { L: ;
# define ENDTRY(x)      } while (0)
#endif


#define SETBITS  32

typedef signed char schar;
typedef unsigned char uchar;

typedef struct {
    void* proc;
	void* link;
} _PROCEDURE;

#ifndef _FNSIZE
# define _FNSIZE  120
#endif


void    PASCAL_MAIN   (int, char **);
char    **P_argv;
int     P_argc;
short   P_escapecode;
int     P_ioresult;
__p2c_jmp_buf *__top_jb;
int     bigendian;

void*   _OutMem     ();
void    _CaseCheck  ();
void    _NilCheck   ();
char   *_ShowEscape   (char *, int, int, char *);
void	_Escape      (int);
void	_EscIO       (int);

int      microsleep    (long);

char    *strsub       (char *, char *, size_t, size_t);
char    *strltrim     (char *);
char    *strrtrim     (char *);
size_t  strpos2      (char *, char *, size_t);
void    strmove      (size_t len, char * s, size_t spos, char * d, size_t dpos);
int     strcicmp     (char * s1, char * s2);

int      P_eof        (FILE *);
int      P_eoln       (FILE *);
long    *P_setunion   (long *, long *, long *) DEPRECATED;
long    *P_setdiff    (long *, long *, long *) DEPRECATED;
int      P_inset      (unsigned, long *) DEPRECATED;
int      P_subset     (long *, long *) DEPRECATED;
long    *P_addset     (long *, unsigned) DEPRECATED;
long    *P_addsetr    (long *, unsigned, unsigned) DEPRECATED;
long    *P_remset     (long *, unsigned) DEPRECATED;
long    *P_expset     (long *, long) DEPRECATED;

char *GetChipmunkPath();

/* I/O error handling */
#define _SETIO(cond,ior)          (P_ioresult = (cond) ? 0 : (ior))

/* Following defines are suitable for the HP Pascal operating system */
#define FileNotFound     10
#define FileNotOpen      13
#define FileWriteError   38
#define BadInputFormat   14
#define EndOfFile        30

/* Creating temporary files */

/* File buffers */
#define FILEBUFNC(f,type)  int f##_BFLAGS;   \
			   type f##_BUFFER

#define RESETBUF(f,type)   (f##_BFLAGS = 1)
#define SETUPBUF(f,type)   (f##_BFLAGS = 0)

#define GETFBUF(f,type)    (*((f##_BFLAGS == 1 &&   \
			       ((f##_BFLAGS = 2),   \
				fread(&f##_BUFFER,  \
				      sizeof(type),1,(f)))),\
			      &f##_BUFFER))

#define GET(f,type)        (f##_BFLAGS == 1 ?   \
			    fread(&f##_BUFFER,sizeof(type),1,(f)) :  \
			    (f##_BFLAGS = 1))

#define PUT(f,type)        (fwrite(&f##_BUFFER,sizeof(type),1,(f)),  \
			    (f##_BFLAGS = 0))

/* Memory allocation */
# define Malloc(n)  (malloc((n != 0) ? (n) : 1) ?: _OutMem())
#define Free(p)     (free((void*)(p)), (p)=NULL)

/* packed arrays */   /* BEWARE: these are untested! */
#define P_getbits_UB(a,i,n,L)   ((int)((a)[(i)>>((L)-(n))] >>   \
				       (((~(i))&((1<<((L)-(n)))-1)) << (n)) &  \
				       ((1<<(1<<(n)))-1)))

#define P_getbits_SB(a,i,n,L)   ((int)((a)[(i)>>((L)-(n))] <<   \
				       (16 - ((((~(i))&((1<<((L)-(n)))-1))+1) <<\
					      (n)) >> (16-(1<<(n))))))

#define P_putbits_UB(a,i,x,n,L) ((a)[(i)>>((L)-(n))] |=   \
				 (x) << (((~(i))&((1<<((L)-(n)))-1)) << (n)))

#define P_putbits_SB(a,i,x,n,L) ((a)[(i)>>((L)-(n))] |=   \
				 ((x) & (1<<(1<<(n)))-1) <<   \
				 (((~(i))&((1<<((L)-(n)))-1)) << (n)))

#define P_clrbits_B(a,i,n,L)    ((a)[(i)>>((L)-(n))] &=   \
				 ~( ((1<<(1<<(n)))-1) <<   \
				   (((~(i))&((1<<((L)-(n)))-1)) << (n))) )

/* small packed arrays */
#define P_getbits_US(v,i,n)     ((int)((v) >> ((i)<<(n)) & (1<<(1<<(n)))-1))
#define P_getbits_SS(v,i,n)     ((int)((long)(v) << (SETBITS - (((i)+1) << (n))) >> (SETBITS-(1<<(n)))))
#define P_putbits_US(v,i,x,n)   ((v) |= (x) << ((i) << (n)))
#define P_putbits_SS(v,i,x,n)   ((v) |= ((x) & (1<<(1<<(n)))-1) << ((i)<<(n)))
#define P_clrbits_S(v,i,n)      ((v) &= ~( ((1<<(1<<(n)))-1) << ((i)<<(n)) ))

#define P_max(a,b)   ((a) > (b) ? (a) : (b))
#define P_min(a,b)   ((a) < (b) ? (a) : (b))

#endif    /* P2C_H */


#if !defined(L_cuserid)
#define L_cuserid 32
char * cuserid(char *);
#endif


