#include "keyboard.h"

#include <graphics/newci.h>
#include "logglobals.h"
#include "logdef.h"
#include "log.h"
#include "window.h"
#include "pen.h"

/// Return TRUE if a key has been pressed (via keyboard or menu boxes).
int pollkbd2()
{
	return (nk_keybufsize() != 0 || pushedbackkey != '\0');
}

/// Return the most recently pressed key.
char testkey2()
{
	if (pushedbackkey != '\0')
		return pushedbackkey;
	else if (pollkbd2())
		return (nk_testkey(0));
	else
		return ' ';
}


/// Return a keystroke. Key is removed from the buffer.
char inkey2()
{
	char ch;

	do
	{
	} while (!pollkbd2());
	if (pushedbackkey != '\0')
	{
		realkey = pushedbackkey;
		pushedbackkey = '\0';
	}
	else
	{
		realkey = nk_getkey();
	}
	ch = realkey;
	if ((ch & 255) >= 168 && (ch & 255) <= 239 && nk_capslock)
	{
		if ((ch & 255) <= 193)
		{
			ch += 46;
		} else if ((ch & 255) >= 214)
		{
			ch -= 46;
		}
	}
	gg.fastspeed = gg.fastmin;
	return ch;
}


void inkey3(char *ch)
{
	*ch = inkey2();
}


char inkey4()
{
	/* undo CAPSLOCK key */
	char ch;

	ch = inkey2();
	if (!nk_capslock)
		return ch;
	if (isupper(ch))
		ch += 32;
	else if (islower(ch))
		ch -= 32;
	return ch;
}


void ungetkey2(char ch)
{
	pushedbackkey = ch;
}

/// Wait until a key is pressed (but ignore it).
void waitforkey()
{
	char ch;

	while (pollkbd2())
		ch = inkey2();
	do
	{
		pen();
	} while (!(pollkbd2() || (gg.t.dn && gg.t.off)));
	if (pollkbd2())
		ch = inkey2();
}

/// Check the keyboard for macros.
void trykbd()
{
	long t;
	int moving;
	char ch;
	macrorec *mp;
	strlist_t *todo;
	char TEMP;

	if (gg.t.near_ && !gg.t0.near_ && gg.showpage >= 0)
	{
		m_graphics_on();
		m_alpha_off();
	}
	if (thingstodo != NULL)
	{
		if (nexttodo == NULL)
			todo = thingstodo;
		else
			todo = nexttodo;
		nexttodo = todo->next;
		assertfunc(todo->s);
		if (nexttodo == NULL)
			strlist_empty(&thingstodo);
		return;
	}
	if (!pollkbd2())
		return;
	norabbits();
	rabtime = timers_sysclock();
	ch = inkey4();
	switch (ch)
	{

		case '\003':
			assertfunc("ABORT");
			break;

		case '\b':
		case '\034':   /* Scroll horizontally */
			do
			{
				m_graphics_on();
				m_colormode((long)m_xor);
				m_color((long)gg.color.scroll);
				m_drawline(xoff0 - 1, 0L, xoff0 - 1, baseline - 1L);
				m_drawline(xoff0 + across + 1, 0L, xoff0 + across + 1, baseline - 1L);
				if (ch == '\b')
					xoff0 -= gg.scale * hscrollspd;
				if (ch == '\034')
					xoff0 += gg.scale * hscrollspd;
				m_drawline(xoff0 - 1, 0L, xoff0 - 1, baseline - 1L);
				m_drawline(xoff0 + across + 1, 0L, xoff0 + across + 1, baseline - 1L);
				m_colormode((long)m_normal);
				t = timers_sysclock() + knobwait;
				do {
				} while (timers_sysclock() <= t && testkey2() == ' ');
				TEMP = testkey2();
				moving = ((uchar)TEMP < 32 && ((1L << TEMP) & 0x10000100L) != 0);
				if (moving)
					ch = inkey2();
			} while (moving && labs(xoff0) <= across - 10);
			refrfunc();
			break;

		case '\n':
		case '\037':   /* Scroll vertically */
			do
			{
				m_graphics_on();
				m_colormode((long)m_xor);
				m_color((long)gg.color.scroll);
				if (yoff0 < baseline)
					m_drawline(0L, yoff0 - 1, (long)across, yoff0 - 1);
				if (yoff0 < 0)
					m_drawline(0L, yoff0 + baseline, (long)across, yoff0 + baseline);
				if (ch == '\n')
					yoff0 += gg.scale * vscrollspd;
				if (ch == '\037')
					yoff0 -= gg.scale * vscrollspd;
				if (yoff0 < baseline)
					m_drawline(0L, yoff0 - 1, (long)across, yoff0 - 1);
				if (yoff0 < 0)
					m_drawline(0L, yoff0 + baseline, (long)across, yoff0 + baseline);
				m_colormode((long)m_normal);
				t = timers_sysclock() + knobwait;
				do
				{
				} while (timers_sysclock() <= t && testkey2() == ' ');
				TEMP = testkey2();
				moving = ((uchar)TEMP < 32 && ((1L << TEMP) & 0x80000400L) != 0);
				if (moving)
					ch = inkey2();
			} while (moving && labs(yoff0) <= baseline - 10);
			refrfunc();
			break;

		default:
			mp = macrobase;
			while (mp != NULL && mp->key != ch)
				mp = mp->next;
			if (mp != NULL)
				assertfunc(mp->name);
			break;
	}
}

void trykbdscroll()
{
	immedscroll = true;
	trykbd();
	immedscroll = false;
	scroll();
}

