
#include <utils/p2c.h>
#include <graphics/newkbd.h>
#include "logglobals.h"
#include "logfont.h"
#include "utils.h"
#include "page.h"
#include "log.h"
#include "box.h"

void chpage(log_page *page)
{
	log_regrec *region;

	stamp(&page->pagestamp);
	page->pagechanged = true;
	region = page->pageregions;
	while (region != NULL)
	{
		stamp(&region->regstamp);
		region = region->next;
	}
}


void chpageplace(log_page *page, short x1, short y1, short x2, short y2)
{
	log_regrec *r;

	stamp(&page->pagestamp);
	page->pagechanged = true;
	r = page->pageregions;
	if (r == NULL)
		return;
	sortshints(&x1, &x2);
	sortshints(&y1, &y2);
	while (r != NULL)
	{
		if (r->x1 <= x2 && x1 <= r->x2 && r->y1 <= y2 && y1 <= r->y2)
			stamp(&r->regstamp);
		r = r->next;
	}
}


int isPageEmpty(log_page *page)
{
	return  page->gbase == NULL && page->hwbase == NULL &&
			page->vwbase == NULL && page->bbase == NULL &&
			page->lbase == NULL;
}


int isPageNull(log_page *page)
{
	return (isPageEmpty(page) && page->pageregions == NULL);
}

int isPageChanged(log_page *page)
{
	return (page->pagechanged && !isPageEmpty(page));
}


int anychanged()
{
	size_t i;

	i = 1;
	while (i <= gg.numpages && !isPageChanged(gg.pages[i - 1]))
		i++;
	return (i <= gg.numpages);
}

/// Switch to a new page.
void newpage(short pg)
{
	short i;

	garbagecoll();
	for (i = gg.numpages; i < pg; i++)
	{
		xoffp[i] = (origin + across / 2) * gg.scale / log_scale0 - across / 2;
		yoffp[i] = (origin + baseline / 2) * gg.scale / log_scale0 - baseline / 2;
		gatecount[i] = 0;
		gg.pages[i]->gbase = NULL;
		gg.pages[i]->sbase = NULL;
		gg.pages[i]->hwbase = NULL;
		gg.pages[i]->vwbase = NULL;
		gg.pages[i]->lbase = NULL;
		gg.pages[i]->bbase = NULL;
		gg.pages[i]->pagechanged = false;
		stamp(&gg.pages[i]->pagestamp);
		gg.pages[i]->pageregions = NULL;
		curfilename[i] = NULL;
		vlsimode[i] = vlsi;
	}
	if (pg != gg.curpage)
	{
		xoffp[gg.curpage - 1] = gg.xoff;
		yoffp[gg.curpage - 1] = gg.yoff;
		gg.xoff = xoffp[pg - 1];
		gg.yoff = yoffp[pg - 1];
		vlsi = vlsimode[pg - 1];
		curwcolor = log_wcol_normal;
		gg.numpages = P_imax2((long)pg, gg.numpages);
		gg.curpage = pg;
		realcurpage = pg;
	}
	while (gg.numpages > gg.curpage && isPageNull(gg.pages[gg.numpages - 1]))
		gg.numpages--;
}

int pagembb(short pg, short *x1, short *y1, short *x2, short *y2)
{
	short max;
	log_grec *g;
	log_hwrec *hw;
	log_vwrec *vw;
	log_lrec *l;
	log_brec *b;

	*x1 = log_maxshint;
	*y1 = log_maxshint;
	*x2 = -log_maxshint;
	*y2 = -log_maxshint;
	g = gg.pages[pg - 1]->gbase;
	while (g != NULL)
	{
		max = g->kind->bbmax;
		if (g->x - max < *x1)
			*x1 = g->x - max;
		if (g->x + max > *x2)
			*x2 = g->x + max;
		if (g->y - max < *y1)
			*y1 = g->y - max;
		if (g->y + max > *y2)
			*y2 = g->y + max;
		g = g->next;
	}
	hw = gg.pages[pg - 1]->hwbase;
	while (hw != NULL)
	{
		if (hw->x1 < *x1)
			*x1 = hw->x1;
		if (hw->x2 > *x2)
			*x2 = hw->x2;
		if (hw->y < *y1)
			*y1 = hw->y;
		if (hw->y > *y2)
			*y2 = hw->y;
		hw = hw->next;
	}
	vw = gg.pages[pg - 1]->vwbase;
	while (vw != NULL)
	{
		if (vw->x < *x1)
			*x1 = vw->x;
		if (vw->x > *x2)
			*x2 = vw->x;
		if (vw->y1 < *y1)
			*y1 = vw->y1;
		if (vw->y2 > *y2)
			*y2 = vw->y2;
		vw = vw->next;
	}
	l = gg.pages[pg - 1]->lbase;
	while (l != NULL)
	{
		if (l->x < *x1)
			*x1 = l->x;
		max = m_strwidth(logfont_lfont, l->name) / gg.scale;
		if (l->x + max > *x2)
			*x2 = l->x + max;
		if (l->y < *y1)
			*y1 = l->y;
		if (l->y + 2 > *y2)
			*y2 = l->y + 2;
		l = l->next;
	}
	b = gg.pages[pg - 1]->bbase;
	while (b != NULL)
	{
		if (b->x1 < *x1)
			*x1 = b->x1;
		if (b->x2 > *x2)
			*x2 = b->x2;
		if (b->y1 < *y1)
			*y1 = b->y1;
		if (b->y2 > *y2)
			*y2 = b->y2;
		b = b->next;
	}
	(*x1)--;
	(*y1)--;
	(*x2)++;
	(*y2)++;
	return (*x1 < *x2);
}

log_page *pageAlloc()
{
	log_page *newpage = Malloc(sizeof(log_page));

	newpage->gbase = NULL;
	newpage->sbase = NULL;
	newpage->hwbase = NULL;
	newpage->vwbase = NULL;
	newpage->lbase = NULL;
	newpage->bbase = NULL;
	newpage->pageregions = NULL;
	newpage->pagechanged = false;
	newpage->pagestamp = 0;

	return newpage;
}

void pageFree(log_page **page)
{
	Free(*page);
	*page = NULL;
}

void pageClear(log_page *page)
{
	log_hwrec *hw;
	log_vwrec *vw;
	log_srec *s;
	log_grec *g;
	log_lrec *l;
	log_brec *b;

	hw = page->hwbase;
	while (hw != NULL)
	{
		page->hwbase = hw->next;
		stamp(&hw->node->simtype->netstamp);
		Free(hw);
		hw = page->hwbase;
	}

	vw = page->vwbase;
	while (vw != NULL)
	{
		page->vwbase = vw->next;
		stamp(&vw->node->simtype->netstamp);
		Free(vw);
		vw = page->vwbase;
	}

	s = page->sbase;
	while (s != NULL)
	{
		page->sbase = s->next;
		Free(s);
		s = page->sbase;
	}

	g = page->gbase;
	while (g != NULL)
	{
		page->gbase = g->next;
		stamp(&g->kind->simtype->netstamp);
		disposegate(&g);
		g = page->gbase;
	}

	l = page->lbase;
	if (l != NULL)
		stamp(&gg.labelstamp);
	while (l != NULL)
	{
		page->lbase = l->next;
		Free(l);
		l = page->lbase;
	}

	b = page->bbase;
	if (b != NULL)
		stamp(&gg.boxstamp);

	while (b != NULL)
	{
		page->bbase = b->next;
		Free(b);
		b = page->bbase;
	}
}

