#include "pagereader.h"

#include <utils/p2c.h>
#include <utils/strings.h>
#include <graphics/newci.h>
#include "logfont.h"
#include "node.h"
#include "log.h"
#include "page.h"
#include "window.h"
#include "tool.h"
#include "utils.h"
#include "box.h"
#include "libraryreader.h"

void loadfail(char *msg, struct LOC_loadpage *LINK)
{
	if (LINK->f)
		fclose(LINK->f);
	if (LINK->ip)
		Free(LINK->ip);
	if (msg == NULL)
		_Escape(0);
	strcpy(LINK->reason, msg);
	_Escape(1);
}

log_nrec *countnode_(short i, struct LOC_loadpage *LINK)
{
	char STR2[256];

	if (i >= 1 && i <= LINK->numnodes)
		return (refnode(LINK->np[i - 1]));
	sprintf(STR2, "Bad node number: %d", i);
	loadfail(STR2, LINK);

	return NULL; /* should never happen */
}

int tryfindfile(char *name, char *path, struct LOC_initialize *LINK)
{
	int result;
	char fname[256];
	FILE *f;
	char *ep;

	f = NULL;
	result = false;
	if (*name != '\0' && (*path == '\0' || (name[0] != '*' && name[0] != '/' && name[0] != '%' && name[0] != '~')))
	{
		TRY(try31);
		if (name[0] == '~')
		{
			ep = tilde_expand(name);
			strncpy(fname, ep, 256);
			Free(ep);
		}
		else
			sprintf(fname, "%s%s", path, name);

		if (f != NULL)
			f = freopen(fname, "r", f);
		else
			f = fopen(fname, "r");

		if (f == NULL)
		{
			P_escapecode = -10;
			P_ioresult = FileNotFound;
			goto _Ltry31;
		}
		strcpy(name, fname);
		result = true;
		RECOVER2(try31,_Ltry31);
		if (P_escapecode == -20)
			_Escape(P_escapecode);
		ENDTRY(try31);
	}
	if (f != NULL)
		fclose(f);
	return result;
}

int locatefile(char *name, struct LOC_initialize *LINK)
{
	char path[256];

	sprintf(path, "%s/", GetChipmunkPath("LOGLIB", LOGLIB));
	return (tryfindfile(name, "", LINK) ||
			tryfindfile(name, gg.homedirname, LINK) ||
			tryfindfile(name, path, LINK));
}


void loadpage(char *filename_, char *reason_)
{
	struct LOC_loadpage V;
	char filename[256];
	log_nrec *n, *n1;
	log_grec *g, *g1;
	log_srec *s;
	log_hwrec *hw;
	log_vwrec *vw;
	log_lrec *l;
	log_brec *b;
	log_hnrec *hn;
	short i, j, k, x, y, x1, y1, x2, y2, numwires, count, ver;
	long ii;
	cnrec *cnbase;
	char tempname[256];
	char labelname[log_lablen + 1];
	char ch, ch1, ch2;
	wirerec *wp;
	char STR2[20];
	char STR3[18];
	short FORLIM;
	char *TEMP;

	strcpy(filename, filename_);
	V.reason = reason_;
	V.f = NULL;
	V.ip = NULL;
	*V.reason = '\0';
	if (*filename != '\0')
	{
		newci_fixfname(filename, "lgf", "");
#ifdef LOAD_SEARCH
		locatefile(filename, NULL);
#endif
		if (V.f != NULL)
			V.f = freopen(filename, "r", V.f);
		else
			V.f = fopen(filename, "r");
		if (V.f == NULL)
			_EscIO(FileNotFound);
		fscanf(V.f, "%hd", &ver);
		ver = -ver;
		gg.actx = gg.curpage - 1;
		calltools(act_clear);
		
		pageClear(gg.pages[gg.curpage - 1]);

		garbagecoll();
		stamp(&gg.loadstamp);
		stamp(&gg.pages[gg.curpage - 1]->pagestamp);
		if (ver < 5)
		{
			loadlog30(&V.f, ver);
		}
		else
		{
			cnbase = NULL;
			fscanf(V.f, "%*[^\n]");
			getc(V.f);
			fscanf(V.f, "%c%c%c%*[^\n]", &ch, &ch1, &ch2);
			getc(V.f);
			if (ch == '\n')
				ch = ' ';
			if (ch1 == '\n')
				ch1 = ' ';
			if (ch2 == '\n')
				ch2 = ' ';
			if (tolower(ch) != 'f' || ch1 != ' ')
				loadfail("Bad format line", &V);

			if (tolower(ch2) != 's')
			{
				sprintf(STR2, "Bad format name: %c", ch2);
				loadfail(STR2, &V);
			}
			vlsi = false;
			setvlsimode(vlsi);
			fscanf(V.f, "%c%hd%*[^\n]", &ch, &V.numnodes);
			getc(V.f);
			if (ch == '\n')
				ch = ' ';
			if (tolower(ch) != 'n')
				loadfail("No node-count line", &V);
			V.np = (log_nrec **)Malloc(V.numnodes * sizeof(log_nrec *));
			n = gg.nbase;
			while (n != NULL && n->next != NULL)
				n = n->next;
			FORLIM = V.numnodes;
			for (i = 0; i < FORLIM; i++)
			{
				fscanf(V.f, "%hd%hd%*[^\n]", &j, &x1);
				getc(V.f);
				n1 = n;
				newnode(&n, j);
				gg.nbase = n->next;
				if (n1 == NULL)
					gg.nbase = n;
				else
					n1->next = n;
				n->next = NULL;
				readattrs(&V.f, &n->attr, n->simtype->nnumattrs, n->simtype->nattr, n->simtype->nlbl, x1);
				V.np[i] = n;
				gg.actfile = &V.f;
				calltoolnode(n, act_readnode);
				stamp(&n->simtype->netstamp);
				while (!P_eoln(V.f) && !P_eof(V.f))
				{
					fscanf(V.f, "%*[^\n]");
					getc(V.f);
				}
				if (P_eof(V.f))
					loadfail("End-of-file while reading nodes", &V);
				fscanf(V.f, "%*[^\n]");
				getc(V.f);
			}
			fscanf(V.f, "%c%hd%*[^\n]", &ch, &count);
			getc(V.f);
			if (ch == '\n')
				ch = ' ';
			if (tolower(ch) != 's')
				loadfail("No signal-count line", &V);
			V.ip = (long *) Malloc(count * sizeof(long));
			for (i = 0; i < count; i++)
			{
				fscanf(V.f, "%hd%c", &j, &ch);
				if (ch == '\n')
					ch = ' ';
				fgets(tempname, 256, V.f);
				TEMP = (char *)strchr(tempname, '\n');
				if (TEMP != NULL)
					*TEMP = 0;
				n = V.np[j - 1];
				x = getsignal(0, tempname);
				if (x == 0)
					loadfail(NULL, &V);
				V.ip[i] = x;
				queuecombine(&cnbase, &V.np[j - 1], &gg.signaltab[x - 1].np);
			}
			fscanf(V.f, "%c%hd%*[^\n]", &ch, &count);
			getc(V.f);
			if (ch == '\n')
				ch = ' ';
			if (tolower(ch) != 'w')
				loadfail("No wire-count line", &V);
			numwires = count;
			wp = (wirerec *)Malloc(numwires * sizeof(wirerec));
			for (i = 0; i < numwires; i++)
			{
				fscanf(V.f, "%hd%hd%hd%hd%hd%c%c%*[^\n]", &x1, &y1, &x2, &y2, &j, &ch,
						&ch1);
				getc(V.f);
				if (ch == '\n')
					ch = ' ';
				if (ch1 == '\n')
					ch1 = ' ';
				switch (ch1)
				{

					case 'U':
						x = log_wcol_normal;
						break;

					case 'G':
						x = log_wcol_green;
						break;

					case 'R':
						x = log_wcol_red;
						break;

					case 'B':
						x = log_wcol_blue;
						break;

					case 'Y':
						x = log_wcol_yellow;
						break;

					default:
						sprintf(STR3, "Bad color name: %c", ch1);
						loadfail(STR3, &V);
						break;
				}

				n = countnode_(j, &V);
				if (y1 == y2)
				{
					newhw(&hw);
					hw->x1 = x1;
					hw->x2 = x2;
					hw->y = y1;
					hw->node = n;
					hw->wcolr = x;
					wp[i].horiz = true;
					wp[i].UU.hw = hw;
				}
				else if (x1 == x2)
				{
					newvw(&vw);
					vw->x = x1;
					vw->y1 = y1;
					vw->y2 = y2;
					vw->node = n;
					vw->wcolr = x;
					wp[i].horiz = false;
					wp[i].UU.vw = vw;
				}
				else
				{
					loadfail("Wire is not horizontal or vertical", &V);
				}
			}
			fscanf(V.f, "%c%hd%*[^\n]", &ch, &count);
			getc(V.f);
			if (ch == '\n')
				ch = ' ';

			if (tolower(ch) != 'p')
				loadfail("No solder-count line", &V);

			for (i = 1; i <= count; i++)
			{
				fscanf(V.f, "%hd%hd", &x, &y);
				newsolder(&s);
				s->x = x;
				s->y = y;
				while (!P_eoln(V.f))
				{
					fscanf(V.f, "%hd", &j);
					if (wp[j - 1].horiz) {
						if (s->hwire == NULL)
							s->hwire = wp[j - 1].UU.hw;
						else
							s->hwire2 = wp[j - 1].UU.hw;
					}
					else
					{
						if (s->vwire == NULL)
							s->vwire = wp[j - 1].UU.vw;
						else
							s->vwire2 = wp[j - 1].UU.vw;
					}
				}
				fscanf(V.f, "%*[^\n]");
				getc(V.f);
			}
			fscanf(V.f, "%c%hd%*[^\n]", &ch, &count);
			getc(V.f);
			if (ch == '\n')
				ch = ' ';
			if (tolower(ch) != 'l')
				loadfail("No label-count line", &V);
			if (count > 0)
				stamp(&gg.labelstamp);

			for (i = 1; i <= count; i++)
			{
				fscanf(V.f, "%hd%hd%hd%c", &x, &y, &j, &ch);
				if (ch == '\n')
					ch = ' ';
				fgets(labelname, log_lablen + 1, V.f);
				TEMP = (char *)strchr(labelname, '\n');
				if (TEMP != NULL)
					*TEMP = 0;
				if (ch != ' ')
					loadfail("Bad label name", &V);
				newlabel(&l);
				l->x = x;
				l->y = y;
				strcpy(l->name, labelname);
				l->w = m_strwidth(logfont_lfont, l->name) / log_scale0;
			}
			fscanf(V.f, "%c%hd%*[^\n]", &ch, &count);
			getc(V.f);
			if (ch == '\n')
				ch = ' ';
			if (tolower(ch) != 'b')
				loadfail("No box-count line", &V);
			if (count > 0)
				stamp(&gg.boxstamp);

			for (i = 1; i <= count; i++)
			{
				fscanf(V.f, "%hd%hd%hd%hd%*[^\n]", &x1, &y1, &x2, &y2);
				getc(V.f);
				newbox(&b);
				b->x1 = x1;
				b->y1 = y1;
				b->x2 = x2;
				b->y2 = y2;
			}

			fscanf(V.f, "%c%hd%*[^\n]", &ch, &count);
			getc(V.f);
			if (ch == '\n')
				ch = ' ';
			if (tolower(ch) != 'g')
				loadfail("No gate-count line", &V);
			g = NULL;
			for (i = 1; i <= count; i++)
			{
				fgets(tempname, 256, V.f);
				TEMP = (char *)strchr(tempname, '\n');
				if (TEMP != NULL)
					*TEMP = 0;
				j = readlibrary(tempname);
				if (j != 0)
				{
					fscanf(V.f, "%hd%hd%hd%hd%ld%hd%hd%*[^\n]", &x, &y, &x1, &y1, &ii,
							&x2, &y2);
					getc(V.f);
					g1 = g;
					newgate(&g, j + x1 * log_kindoffset);
					if (g1 == NULL)
						gg.pages[gg.curpage - 1]->gbase = g;
					else
						g1->next = g;
					g->next = NULL;
					g->rot = x1;
					g->x = x;
					g->y = y;
					g->gc = y1;
					g->vars = (na_long)ii;
					if (x2 == 0)
						g->sig = 0;
					else
						g->sig = V.ip[x2 - 1];
					stamp(&g->kind->simtype->netstamp);
					j = 1;
					while (j <= g->kind->numpins && !P_eoln(V.f))
					{
						fscanf(V.f, "%hd", &k);
						g->pin[j - 1] = countnode_(k, &V);
						j++;
					}
					while (j <= g->kind->numpins)
					{
						newnode(&g->pin[j - 1], g->kind->pin[j - 1].s);
						j++;
					}
					fscanf(V.f, "%*[^\n]");
					getc(V.f);
					readattrs(&V.f, &g->attr, g->kind->numattrs, g->kind->attr, g->kind->lbl, y2);
					initpinpos(g);
					gg.actflag = true;
					calltoolgate(g, act_connectgate);
					if (!gg.actflag)
						report(11, "LOADPAGE");
					gg.actfile = &V.f;
					gg.actflag = true;
					calltoolgate(g, act_readgate);
				}
				while (!P_eoln(V.f) && !P_eof(V.f))
				{
					fscanf(V.f, "%*[^\n]");
					getc(V.f);
				}
				if (P_eof(V.f))
					loadfail("End-of-file while reading gates", &V);
				fscanf(V.f, "%*[^\n]");
				getc(V.f);
			}
			if (!checkcombine(&cnbase))
				loadfail("Simtype collision on signal names", &V);
			ch = getc(V.f);
			if (ch == '\n')
				ch = ' ';
			if (tolower(ch) == 'h')
			{
				fscanf(V.f, "%hd%*[^\n]", &count);
				getc(V.f);
				histdelsignals();
				for (i = 1; i <= count; i++)
				{
					fscanf(V.f, "%hd%hd%c", &y, &x1, &ch);
					if (ch == '\n')
						ch = ' ';
					fgets(tempname, 256, V.f);
					TEMP = (char *)strchr(tempname, '\n');
					if (TEMP != NULL)
						*TEMP = 0;
					x = getsignal(0, tempname);
					if (x == 0)
						loadfail(NULL, &V);
					n = gg.signaltab[x - 1].np;
					histaddsignal(&hn, x, y);
					readattrs(&V.f, &hn->attr, n->simtype->hnumattrs, n->simtype->hattr,
							n->simtype->hlbl, x1);
				}
				ch = getc(V.f);
				if (ch == '\n')
					ch = ' ';
			}
			if (tolower(ch) != '.')
				loadfail("No end line in file", &V);
		}
		if (V.f != NULL)
			fclose(V.f);
		V.f = NULL;
		if (V.ip != NULL)
			Free(V.ip);
		chpage(gg.pages[gg.curpage - 1]);
		gg.pages[gg.curpage - 1]->pagechanged = false;
		if (curfilename[gg.curpage - 1] == NULL)
			curfilename[gg.curpage - 1] = (char *)Malloc(256);
		strcpy(curfilename[gg.curpage - 1], filename);
	}
}

void loadfail_(char *msg, struct LOC_readlgf *LINK)
{
	if (*LINK->f)
		fclose(*LINK->f);
	if (LINK->ip)
		Free(LINK->ip);
	if (msg == NULL)
		_Escape(0);
	strcpy(LINK->reason, msg);
	_Escape(1);
}


static void readlgf(FILE **f, char *reason_)
{
	struct LOC_readlgf V;
	char buf[256];
	char ch;
	long i, j, k, m, n, ii;
	short x1, y1, x2, y2;
	log_grec *g;
	log_hnrec *hn;
	log_nrec *n1;
	char *TEMP;
	char STR2[38];

	V.reason = reason_;
	V.f = f;
	V.ip = NULL;
	fgets(buf, 256, *V.f);
	TEMP = (char *)strchr(buf, '\n');
	if (TEMP != NULL)
		*TEMP = 0;
	if (strcmp(buf, "f s"))
		loadfail_("Not an LGF file", &V);
	do
	{
		while (P_eoln(*V.f) && !P_eof(*V.f))
		{
			fscanf(*V.f, "%*[^\n]");
			getc(*V.f);
		}
		do
		{
			if (P_eof(*V.f))
				ch = '.';
			else {
				ch = getc(*V.f);
				if (ch == '\n')
					ch = ' ';
			}
		} while (ch == ' ');

		switch (tolower(ch))
		{

			case '#':   /*comment*/
				fscanf(*V.f, "%*[^\n]");
				getc(*V.f);
				break;

			case 'n':   /*nodes*/
				fscanf(*V.f, "%ld%*[^\n]", &i);
				getc(*V.f);
				for (j = 1; j <= i; j++) {
					fscanf(*V.f, "%ld%ld%*[^\n]", &m, &n);
					getc(*V.f);   /*simtype, num attrs*/
					for (m = 1; m <= n; m++)
					{   /*skip attributes*/
						fscanf(*V.f, "%*[^\n]");
						getc(*V.f);
					}
					while (!P_eoln(*V.f) && !P_eof(*V.f))
					{   /*skip simtype info*/
						fscanf(*V.f, "%*[^\n]");
						getc(*V.f);
					}
				}
				break;

			case 's':   /*signal names*/
				fscanf(*V.f, "%ld%*[^\n]", &i);
				getc(*V.f);
				V.ip = (long *)Malloc(i * sizeof(long) );
				for (j = 0; j < i; j++)
				{
					fscanf(*V.f, "%ld%c", &n, &ch);
					if (ch == '\n')
						ch = ' ';
					fgets(buf, 256, *V.f);
					TEMP = (char *)strchr(buf, '\n');
					if (TEMP != NULL)
						*TEMP = 0;
					V.ip[j] = getsignal(0, buf);
					if (V.ip[j] == 0)
						loadfail_(NULL, &V);
				}
				break;

			case 'w':   /*wires*/
				fscanf(*V.f, "%ld%*[^\n]", &i);
				getc(*V.f);
				for (j = 1; j <= i; j++)
				{
					fscanf(*V.f, "%hd%hd%hd%hd%*[^\n]", &x1, &y1, &x2, &y2);
					getc(*V.f);   /*skip node & color info*/
					if (x1 != x2)
						addhwire2(x1, x2, y1);
					else if (y1 != y2)
						addvwire2(x1, y1, y2);
				}
				break;

			case 'p':   /*solder*/
				fscanf(*V.f, "%ld%*[^\n]", &i);
				getc(*V.f);
				for (j = 1; j <= i; j++)
				{
					fscanf(*V.f, "%hd%hd%*[^\n]", &x1, &y1);
					getc(*V.f);   /*skip over wire numbers*/
					solderat(x1, y1);
				}
				break;

			case 'l':   /*labels*/
				fscanf(*V.f, "%ld%*[^\n]", &i);
				getc(*V.f);
				for (j = 1; j <= i; j++)
				{
					fscanf(*V.f, "%hd%hd%ld%c", &x1, &y1, &k, &ch);
					if (ch == '\n')
						ch = ' ';
					fgets(buf, 256, *V.f);
					TEMP = (char *) strchr(buf, '\n');
					if (TEMP != NULL)
						*TEMP = 0;
					if (ch != ' ')
						loadfail_("Bad label name", &V);
					addlabelat(x1, y1, buf);
				}
				break;

			case 'b':   /*boxes*/
				fscanf(*V.f, "%ld%*[^\n]", &i);
				getc(*V.f);
				for (j = 1; j <= i; j++)
				{
					fscanf(*V.f, "%hd%hd%hd%hd%*[^\n]", &x1, &y1, &x2, &y2);
					getc(*V.f);
					addboxat(x1, y1, x2, y2);
				}
				break;

			case 'g':   /*gates*/
				fscanf(*V.f, "%ld%*[^\n]", &i);
				getc(*V.f);
				for (j = 1; j <= i; j++)
				{
					fgets(buf, 256, *V.f);
					TEMP = (char *)strchr(buf, '\n');
					if (TEMP != NULL)
						*TEMP = 0;
					k = readlibrary(buf);
					if (k != 0) {
						fscanf(*V.f, "%hd%hd%hd%hd%ld%ld%ld%*[^\n]", &x1, &y1, &x2, &y2, &ii, &m, &n);
						getc(*V.f);
						addgate1(x1, y1, (int)(k + x2 * log_kindoffset));
						if (gg.neargate != NULL)
						{
							g = gg.neargate;
							g->gc = y2;
							g->vars = (na_long)ii;
							if (m != 0)
								settofrom(&g, gg.signaltab[ V.ip[m - 1] - 1].name);
							fscanf(*V.f, "%*[^\n]");   /*ignore pins*/
							getc(*V.f);
							readattrs(V.f, &g->attr, g->kind->numattrs, g->kind->attr,
									g->kind->lbl, (int)n);
							gg.actfile = V.f;
							gg.actflag = false;
							calltoolgate(g, act_readgate);
						}
					}
					while (!P_eoln(*V.f) && !P_eof(*V.f))
					{
						fscanf(*V.f, "%*[^\n]");
						getc(*V.f);
					}

					if (P_eof(*V.f))
						loadfail_("End-of-file while reading gates", &V);
					fscanf(*V.f, "%*[^\n]");
					getc(*V.f);
				}
				break;

			case 'h':   /*history names*/
				fscanf(*V.f, "%ld%*[^\n]", &i);
				getc(*V.f);
				histdelsignals();
				for (j = 1; j <= i; j++)
				{
					fscanf(*V.f, "%hd%hd%c", &y1, &x1, &ch);
					if (ch == '\n')
						ch = ' ';
					fgets(buf, 256, *V.f);
					TEMP = (char *)strchr(buf, '\n');
					if (TEMP != NULL)
						*TEMP = 0;
					x2 = getsignal(0, buf);
					if (x2 == 0)
						loadfail_(NULL, &V);
					n1 = gg.signaltab[x2 - 1].np;
					histaddsignal(&hn, x2, y1);
					readattrs(V.f, &hn->attr, n1->simtype->hnumattrs, n1->simtype->hattr, n1->simtype->hlbl, x1);
				}
				break;

			case '.':
				/* blank case */
				break;

			default:
				sprintf(STR2, "Don't understand header character: %c", ch);
				loadfail_(STR2, &V);
				break;
		}
	} while (ch != '.');
	if (V.ip != NULL)
		Free(V.ip);
}


void loadfail__(char *msg, char *reason)
{
	strcpy(reason, msg);
	_Escape(1);
}


void readpage(char *filename_, char *reason)
{
	char filename[256];
	FILE *f;
	char buf[256];
	char *TEMP;

	strcpy(filename, filename_);
	f = NULL;
	*reason = '\0';
	newci_fixfname(filename, "lgf", "");
	if (f != NULL)
		f = freopen(filename, "r", f);
	else
		f = fopen(filename, "r");
	if (f == NULL)
		_EscIO(FileNotFound);
	fgets(buf, 256, f);
	TEMP = (char *)strchr(buf, '\n');
	if (TEMP != NULL)
		*TEMP = 0;
	if (!strcmp(buf, "-5"))
		readlgf(&f, reason);
	else if (*buf == '-')
		loadfail__("Unable to read old-format files", reason);
	else
		loadfail__("Not an LGF file", reason);
	if (f != NULL)
		fclose(f);
	f = NULL;
	chpage(gg.pages[gg.curpage - 1]);
	gg.pages[gg.curpage - 1]->pagechanged = false;
	if (curfilename[gg.curpage - 1] == NULL)
		curfilename[gg.curpage - 1] = (char *)Malloc(256);
	strcpy(curfilename[gg.curpage - 1], filename);
	if (f != NULL)
		fclose(f);
}

void readattrs(FILE **f, log_gattrrec **gattr, short numattrs, log_kattrrec *kattr, strlist_t *lbl, short count)
{
	short j, jj;
	char ch;
	char buf[256];
	strlist_t *l1, *names;

	l1 = lbl;
	while (l1 != NULL && l1->kind != '\001')
		l1 = l1->next;

	if (l1 != NULL)
		names = (strlist_t *)l1->value;
	else
		names = NULL;

	for (j = 1; j <= count; j++)
	{
		ch = getc(*f);
		if (ch == '\n')
			ch = ' ';

		if (ch == '[')
		{
			*buf = '\0';
			ch = getc(*f);

			if (ch == '\n')
				ch = ' ';

			while (ch != ']' && !P_eoln(*f))
			{
				sprintf(buf + strlen(buf), "%c", ch);
				ch = getc(*f);
				if (ch == '\n')
					ch = ' ';
			}

			jj = -1;
			if (!P_eoln(*f))
			{
				l1 = names;
				while (l1 != NULL && strcmp(buf, l1->s))
					l1 = l1->next;

				if (l1 != NULL)
					jj = (long)l1->value;

				do
				{
					ch = getc(*f);
					if (ch == '\n')
						ch = ' ';
				} while (ch == ' ');
			}
		}
		else
		{
			jj = j;
		}

		if (jj >= 1 && jj <= numattrs)
		{
			ch = toupper(ch);
			if (ch == 'X')
			{
				(*gattr)[jj - 1].blnk = true;
				switch (kattr[jj - 1].dtype)
				{

					case 'C':
						*(*gattr)[jj - 1].UU.c = '\0';
						break;

					case 'A':
						strchange(&(*gattr)[jj - 1].UU.sp, "");
						break;
				}
			}
			else
			{
				(*gattr)[jj - 1].blnk = false;
				switch (kattr[jj - 1].dtype)
				{

					case 'R':
					case 'U':
					case 'F':
						if (ch == 'F' || ch == 'U' || ch == 'R')
							fscanf(*f, "%lg", &(*gattr)[jj - 1].UU.r);
						break;

					case 'I':
					case 'H':
						if (ch == 'H' || ch == 'I')
							fscanf(*f, "%ld", &(*gattr)[jj - 1].UU.U73.i1);
						break;

					case 'C':
						if (ch == 'A' || ch == 'C')
						{
							*buf = '\0';
							fscanf(*f, "%[^\n]", buf);
							if (strlen(buf) <= kattr[jj - 1].prec)
								strcpy((*gattr)[jj - 1].UU.c, buf);
						}
						break;

					case 'A':
						if (ch == 'A' || ch == 'C')
						{
							*buf = '\0';
							fscanf(*f, "%[^\n]", buf);
							strchange(&(*gattr)[jj - 1].UU.sp, buf);
						}
						break;

					case 'B':
						if (ch == 'B')
						{
							ch = getc(*f);
							if (ch == '\n')
								ch = ' ';
							(*gattr)[jj - 1].UU.b = (ch == 't' || ch == 'T' || ch == '1');
						}
						break;

					case 'V':
						if (ch == 'V')
						{
							fscanf(*f, "%hd", &(*gattr)[jj - 1].UU.nv);
							if ((*gattr)[jj - 1].UU.nv >= kattr[jj - 1].UU.U86.nv)
								(*gattr)[jj - 1].UU.nv = 0;
						}
						break;
				}
			}
		}
		fscanf(*f, "%*[^\n]");
		getc(*f);
	}
}

void loadlog30(FILE **f, short ver)
{
	short i, j, x, y, x1, y1, x2, y2;
	char nam[81];
	char ch;
	log_lrec *l;
	log_brec *b;
	char *TEMP;

	fscanf(*f, "%*[^\n]");   /*version number*/
	getc(*f);
	fscanf(*f, "%c%*[^\n]", &ch);
	getc(*f);
	if (ch == '\n')
		ch = ' ';
	switch (ch)
	{

		case 'V':
		case 'v':
			vlsi = true;
			break;

		case 'T':
		case 't':
			vlsi = false;
			break;
	}
	fscanf(*f, "%*[^\n]");   /*numnodes, nullnode*/
	getc(*f);
	setvlsimode(vlsi);
	j = gg.color.wire[0];
	if (vlsi)
	{
		fscanf(*f, "%hd%hd%hd%hd%hd%*[^\n]", &x1, &x2, &y, &i, &j);
		getc(*f);
	}
	else
	{
		fscanf(*f, "%hd%hd%hd%hd%*[^\n]", &x1, &x2, &y, &i);
		getc(*f);
	}

	while (x1 != -9999)
	{
		addhwire(x1, x2, y, log_wcol_normal);
		if (vlsi)
		{
			fscanf(*f, "%hd%hd%hd%hd%hd%*[^\n]", &x1, &x2, &y, &i, &j);
			getc(*f);
		}
		else
		{
			fscanf(*f, "%hd%hd%hd%hd%*[^\n]", &x1, &x2, &y, &i);
			getc(*f);
		}
	}

	if (vlsi)
	{
		fscanf(*f, "%hd%hd%hd%hd%hd%*[^\n]", &x, &y1, &y2, &i, &j);
		getc(*f);
	}
	else
	{
		fscanf(*f, "%hd%hd%hd%hd%*[^\n]", &x, &y1, &y2, &i);
		getc(*f);
	}

	while (x != -9999)
	{
		addvwire(x, y1, y2, log_wcol_normal);
		if (vlsi)
		{
			fscanf(*f, "%hd%hd%hd%hd%hd%*[^\n]", &x, &y1, &y2, &i, &j);
			getc(*f);
		}
		else
		{
			fscanf(*f, "%hd%hd%hd%hd%*[^\n]", &x, &y1, &y2, &i);
			getc(*f);
		}
	}
	fscanf(*f, "%hd%hd%hd%hd%*[^\n]", &x, &y, &i, &j);
	getc(*f);
	while (x != -9999)
	{
		closerwire(x, y);
		if (gg.nearhw != NULL && gg.nearvw != NULL)
		{
			if (trycombinenodes(&gg.nearvw->node, &gg.nearhw->node))
			{
				clipon();
				addsolder(x, y, gg.nearhw, NULL, gg.nearvw, NULL);
				clipoff();
			}
			else
			{
				frysolder(x, y);
			}
		}
		fscanf(*f, "%hd%hd%hd%hd%*[^\n]", &x, &y, &i, &j);
		getc(*f);
	}

	if (ver >= 3)
	{
		fscanf(*f, "%hd%hd%hd%c", &x, &y, &i, &ch);
		if (ch == '\n')
			ch = ' ';
		fgets(nam, 81, *f);
		TEMP = (char *)strchr(nam, '\n');
		if (TEMP != NULL)
			*TEMP = 0;
	}
	else
	{
		fscanf(*f, "%hd%hd%c", &x, &y, &ch);
		if (ch == '\n')
			ch = ' ';
		fgets(nam, 81, *f);
		TEMP = (char *)strchr(nam, '\n');
		if (TEMP != NULL)
			*TEMP = 0;
	}

	while (x != -9999)
	{
		newlabel(&l);
		l->x = x;
		l->y = y;
		strcpy(l->name, nam);
		l->w = m_strwidth(logfont_lfont, l->name) / log_scale0;
		if (ver >= 3)
		{
			fscanf(*f, "%hd%hd%hd%c", &x, &y, &i, &ch);
			if (ch == '\n')
				ch = ' ';
			fgets(nam, 81, *f);
			TEMP = (char *)strchr(nam, '\n');
			if (TEMP != NULL)
				*TEMP = 0;
		}
		else
		{
			fscanf(*f, "%hd%hd%c", &x, &y, &ch);
			if (ch == '\n')
				ch = ' ';
			fgets(nam, 81, *f);
			TEMP = (char *)strchr(nam, '\n');
			if (TEMP != NULL)
				*TEMP = 0;
		}
		if (ch != ':')
			_Escape(0);
	}
	fscanf(*f, "%hd%hd%hd%hd%*[^\n]", &x, &y, &x1, &y1);
	getc(*f);
	while (x != -9999)
	{
		newbox(&b);
		b->x1 = x;
		b->y1 = y;
		b->x2 = x1;
		b->y2 = y1;
		fscanf(*f, "%hd%hd%hd%hd%*[^\n]", &x, &y, &x1, &y1);
		getc(*f);
	}
	fscanf(*f, "%hd%hd%hd%hd%hd%hd%hd%c", &i, &x, &y, &y1, &x1, &y2, &x2, &ch);

	if (ch == '\n')
		ch = ' ';

	fgets(nam, 81, *f);
	TEMP = (char *)strchr(nam, '\n');
	if (TEMP != NULL)
		*TEMP = 0;

	while (i != -9999)
	{
		if ((unsigned)(i & 127) < 32 && ((1L << (i & 127)) & 0x1800) != 0)
		{
			if (i == 11)
				j = readlibrary("TO");
			else
				j = readlibrary("FROM");
			addgate1(x, y, j + i / 128 * 128);
			gg.neargate->sig = getsignal(0, nam);
		}
		else
		{
			j = readlibrary(nam);
			addgate1(x, y, j + i / 128 * 128);
		}

		for (i = 1; i <= y2 + x2; i++)
		{   /*ignore all pins, attributes!*/
			fscanf(*f, "%*[^\n]");
			getc(*f);
		}

		fscanf(*f, "%hd%hd%hd%hd%hd%hd%hd%c", &i, &x, &y, &y1, &x1, &y2, &x2, &ch);
		if (ch == '\n')
			ch = ' ';
		fgets(nam, 81, *f);
		TEMP = (char *)strchr(nam, '\n');
		if (TEMP != NULL)
			*TEMP = 0;
	}
	chpage(gg.pages[gg.curpage - 1]);
}
